from __future__ import annotations

import pytest

from ._router_helpers import (
    DemoRoleORM,
    DemoUserORM,
    build_test_app,
    get_test_client,
    save_entities,
)

pytestmark = pytest.mark.asyncio


async def test_router_returns_list_filters_in_model_meta() -> None:
    app, _session_factory = await build_test_app()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/")

        assert response.status_code == 200
        assert response.json()["list_filters"] == [
            {
                "slug": "status",
                "label": "Status",
                "group": "Flags",
                "field_name": "is_active",
                "input_kind": "boolean",
                "placeholder": None,
                "has_choices": True,
                "options": [{"value": "true", "label": "Да"}, {"value": "false", "label": "Нет"}],
            },
            {
                "slug": "role_id",
                "label": "Role",
                "group": "Access",
                "field_name": "roles",
                "input_kind": "select",
                "placeholder": None,
                "has_choices": True,
                "options": [],
            },
            {
                "slug": "username_contains",
                "label": "Username",
                "group": None,
                "field_name": "username",
                "input_kind": "text",
                "placeholder": None,
                "has_choices": False,
                "options": [],
            },
            {
                "slug": "username_prefix",
                "label": "Username Prefix",
                "group": None,
                "field_name": None,
                "input_kind": "text",
                "placeholder": None,
                "has_choices": False,
                "options": [],
            },
            {
                "slug": "audience",
                "label": "Audience",
                "group": "Flags",
                "field_name": None,
                "input_kind": "select",
                "placeholder": None,
                "has_choices": True,
                "options": [
                    {"value": "active-only", "label": "Active only"},
                    {"value": "inactive-only", "label": "Inactive only"},
                ],
            },
        ]


async def test_router_supports_boolean_list_filter() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1", is_active=True),
        DemoUserORM(username="beta", password="hashed::2", is_active=False),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"status": "false"})

        assert response.status_code == 200
        assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: beta"]


async def test_router_returns_relation_filter_choices() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoRoleORM(name="admin"),
        DemoRoleORM(name="manager"),
        DemoRoleORM(name="hidden"),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/filters/role_id/choices/")

        assert response.status_code == 200
        assert response.json()["items"] == [{"id": 1, "label": "admin"}, {"id": 2, "label": "manager"}]


async def test_router_supports_many_to_many_relation_list_filter() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        admin_role = DemoRoleORM(name="admin")
        manager_role = DemoRoleORM(name="manager")
        first_user = DemoUserORM(username="alpha", password="hashed::1", roles=[admin_role])
        second_user = DemoUserORM(username="beta", password="hashed::2", roles=[manager_role])
        session.add_all([first_user, second_user])
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"role_id": "1"})

        assert response.status_code == 200
        assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 1: alpha"]


async def test_router_supports_text_list_filter() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1"),
        DemoUserORM(username="xlar", password="hashed::2"),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"username_contains": "xla"})

        assert response.status_code == 200
        assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: xlar"]


async def test_router_supports_custom_list_filter_handler() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1"),
        DemoUserORM(username="xlar", password="hashed::2"),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"username_prefix": "xla"})

        assert response.status_code == 200
        assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: xlar"]


async def test_router_supports_select_option_specific_filter_handler() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1", is_active=True),
        DemoUserORM(username="beta", password="hashed::2", is_active=False),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"audience": "inactive-only"})

        assert response.status_code == 200
        assert [item["display_name"] for item in response.json()["items"]] == ["Пользователь 2: beta"]


async def test_router_reuses_related_model_search_for_choices() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoRoleORM(name="admin"),
        DemoRoleORM(name="moderator"),
        DemoRoleORM(name="user"),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/audits/fields/role_id/choices/", params={"q": "ad"})

        assert response.status_code == 200
        assert response.json()["items"] == [{"id": 1, "label": "admin"}]


async def test_router_applies_relation_scope_to_field_choices() -> None:
    app, session_factory = await build_test_app()
    await save_entities(session_factory, DemoRoleORM(name="admin"), DemoRoleORM(name="hidden"))

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/audits/fields/role_id/choices/")

        assert response.status_code == 200
        assert response.json()["items"] == [{"id": 1, "label": "admin"}]



from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from typing import Any
from uuid import UUID

from sqlalchemy import inspect as sa_inspect

from xladmin.config import AdminModelConfig
from xladmin.introspection_fields import serialize_related_pk


def convert_value_for_column(column: Any, value: Any) -> Any:
    if value is None:
        return None

    try:
        python_type = column.type.python_type
    except NotImplementedError:
        python_type = None

    if python_type is None:
        return value
    if python_type is bool:
        return convert_boolean_scalar(value)
    if python_type is int:
        return int(value)
    if python_type is float:
        return float(value)
    if python_type is Decimal:
        return Decimal(str(value))
    if python_type is date:
        return date.fromisoformat(str(value))
    if python_type is datetime:
        return datetime.fromisoformat(str(value))
    if python_type is UUID:
        return UUID(str(value))
    if python_type is str:
        return str(value)
    return value


def convert_boolean_scalar(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float)):
        return value != 0
    normalized_value = str(value).strip().lower()
    if normalized_value in {"1", "true", "yes", "y", "on"}:
        return True
    if normalized_value in {"0", "false", "no", "n", "off", ""}:
        return False
    raise ValueError(f"Cannot convert value '{value}' to boolean.")


def get_field_value(config: AdminModelConfig, instance: Any, field_name: str) -> Any:
    field_config = config.get_field_config(field_name)
    if field_config.value_getter is not None:
        return field_config.value_getter(instance)

    mapper = sa_inspect(config.model)
    if field_name in mapper.relationships:
        relationship = mapper.relationships[field_name]
        related_value = getattr(instance, field_name, None)
        if relationship.uselist:
            return [serialize_related_pk(item) for item in list(related_value or [])]
        if related_value is None:
            return None
        return serialize_related_pk(related_value)

    return getattr(instance, field_name, None)



from __future__ import annotations

import pytest
from sqlalchemy import select

from ._router_helpers import (
    DemoAliasedChatORM,
    DemoAliasedMessageORM,
    DemoBadgeORM,
    DemoCommentORM,
    DemoHardChildORM,
    DemoHardParentORM,
    DemoOwnedBoardORM,
    DemoOwnedProjectORM,
    DemoPostORM,
    DemoSecretORM,
    DemoSoftNoteORM,
    DemoSoftUserORM,
    DemoTaskORM,
    DemoTokenProfileORM,
    DemoUserORM,
    build_test_app,
    get_test_client,
)

pytestmark = pytest.mark.asyncio


async def _seed_owned_project_tree(session_factory) -> None:
    async with session_factory() as session:
        user = DemoUserORM(username="owner", password="hashed::1")
        session.add(user)
        await session.flush()
        project = DemoOwnedProjectORM(title="Tracker Project", author_id=user.id)
        session.add(project)
        await session.flush()
        session.add(DemoOwnedBoardORM(title="Roadmap", project_id=project.id))
        await session.commit()


async def test_router_builds_delete_preview_tree_for_object() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        post = DemoPostORM(title="Welcome", user=user)
        post.comments.append(DemoCommentORM(body="First"))
        session.add(user)
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 1, "delete": 2, "protect": 0, "set_null": 0, "total": 3}
        assert payload["roots"][0]["model_slug"] == "users"
        assert payload["roots"][0]["children"][0]["model_slug"] == "posts"
        assert payload["roots"][0]["children"][0]["effect"] == "delete"
        assert payload["roots"][0]["children"][0]["children"][0]["model_slug"] == "comments"


async def test_router_builds_delete_preview_tree_for_bulk_delete() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        first_user = DemoUserORM(username="alpha", password="hashed::1")
        first_post = DemoPostORM(title="Alpha post", user=first_user)
        first_post.comments.append(DemoCommentORM(body="Alpha comment"))
        second_user = DemoUserORM(username="beta", password="hashed::2")
        second_post = DemoPostORM(title="Beta post", user=second_user)
        second_post.comments.append(DemoCommentORM(body="Beta comment"))
        session.add_all([first_user, second_user])
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": [1, 2]})

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 2, "delete": 4, "protect": 0, "set_null": 0, "total": 6}
        assert [root["id"] for root in payload["roots"]] == [1, 2]


async def test_router_returns_empty_bulk_delete_preview_for_empty_ids() -> None:
    app, _session_factory = await build_test_app()

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": []})

        assert response.status_code == 200
        assert response.json() == {
            "can_delete": True,
            "summary": {"roots": 0, "delete": 0, "protect": 0, "set_null": 0, "total": 0},
            "roots": [],
        }


async def test_router_builds_bulk_delete_preview_only_for_existing_ids() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        session.add(DemoUserORM(username="alpha", password="hashed::1"))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/bulk-delete-preview/", json={"ids": [1, 999]})

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 0, "set_null": 0, "total": 1}
        assert [root["id"] for root in payload["roots"]] == [1]


async def test_router_marks_non_cascade_relations_as_protected_in_delete_preview() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        badge = DemoBadgeORM(title="VIP", user=user)
        session.add_all([user, badge])
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is False
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
        assert payload["roots"][0]["children"][0]["model_slug"] == "badges"
        assert payload["roots"][0]["children"][0]["effect"] == "protect"


async def test_router_marks_required_non_cascade_relations_as_protected_in_delete_preview() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.tasks.append(DemoTaskORM(title="Follow-up"))
        session.add(user)
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is False
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
        assert payload["roots"][0]["children"][0]["model_slug"] == "tasks"
        assert payload["roots"][0]["children"][0]["effect"] == "protect"


async def test_router_blocks_delete_for_required_non_cascade_relations() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.tasks.append(DemoTaskORM(title="Follow-up"))
        session.add(user)
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.delete("/xladmin/models/users/items/1/")

        assert response.status_code == 409

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoUserORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoTaskORM))).scalars())) == 1


async def test_router_detects_reverse_foreign_keys_without_parent_relationships() -> None:
    app, session_factory = await build_test_app()
    await _seed_owned_project_tree(session_factory)

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is False
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 2, "set_null": 0, "total": 3}
        project_node = payload["roots"][0]["children"][0]
        assert project_node["model_slug"] == "owned-projects"
        assert project_node["relation_name"] == "author_id"
        assert project_node["effect"] == "protect"
        assert project_node["children"][0]["model_slug"] == "owned-boards"
        assert project_node["children"][0]["relation_name"] == "project_id"
        assert project_node["children"][0]["effect"] == "protect"


async def test_router_blocks_delete_for_reverse_foreign_keys_without_parent_relationships() -> None:
    app, session_factory = await build_test_app()
    await _seed_owned_project_tree(session_factory)

    async with get_test_client(app) as client:
        response = await client.delete("/xladmin/models/users/items/1/")

        assert response.status_code == 409

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoUserORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoOwnedProjectORM))).scalars())) == 1
        assert len(list((await session.execute(select(DemoOwnedBoardORM))).scalars())) == 1


async def test_router_handles_related_models_with_non_default_primary_key() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="owner", password="hashed::1")
        session.add(user)
        await session.flush()
        session.add(DemoTokenProfileORM(tokenprofile_ptr_id="profile-1", title="Profile", user_id=user.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is False
        related_slugs = [child["model_slug"] for child in payload["roots"][0]["children"]]
        assert "token-profiles" in related_slugs


async def test_router_handles_related_models_with_aliased_column_name() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        message = DemoAliasedMessageORM(id="message-1")
        session.add(message)
        await session.flush()
        session.add(DemoAliasedChatORM(id="chat-1", last_branch_message_id=message.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/aliased-messages/items/message-1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is False
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 1, "set_null": 0, "total": 2}
        assert payload["roots"][0]["children"][0]["model_slug"] == "aliased-chats"


async def test_router_marks_ondelete_set_null_relations_in_delete_preview() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoSoftUserORM(username="alpha")
        session.add(user)
        await session.flush()
        session.add(DemoSoftNoteORM(body="note", user_id=user.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/soft-users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 1, "delete": 0, "protect": 0, "set_null": 1, "total": 2}
        assert payload["roots"][0]["children"][0]["model_slug"] == "soft-notes"
        assert payload["roots"][0]["children"][0]["effect"] == "set-null"


async def test_router_applies_set_null_before_delete() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoSoftUserORM(username="alpha")
        session.add(user)
        await session.flush()
        session.add(DemoSoftNoteORM(body="note", user_id=user.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.delete("/xladmin/models/soft-users/items/1/")

        assert response.status_code == 204

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoSoftUserORM))).scalars())) == 0
        note = (await session.execute(select(DemoSoftNoteORM))).scalar_one()
        assert note.user_id is None


async def test_router_marks_ondelete_cascade_relations_as_delete() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        parent = DemoHardParentORM(title="parent")
        session.add(parent)
        await session.flush()
        session.add(DemoHardChildORM(title="child", parent_id=parent.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/hard-parents/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 1, "delete": 1, "protect": 0, "set_null": 0, "total": 2}
        assert payload["roots"][0]["children"][0]["model_slug"] == "hard-children"
        assert payload["roots"][0]["children"][0]["effect"] == "delete"


async def test_router_deletes_ondelete_cascade_relations() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        parent = DemoHardParentORM(title="parent")
        session.add(parent)
        await session.flush()
        session.add(DemoHardChildORM(title="child", parent_id=parent.id))
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.delete("/xladmin/models/hard-parents/items/1/")

        assert response.status_code == 204

    async with session_factory() as session:
        assert len(list((await session.execute(select(DemoHardParentORM))).scalars())) == 0
        assert len(list((await session.execute(select(DemoHardChildORM))).scalars())) == 0


async def test_router_builds_delete_preview_with_multiple_cascade_branches() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        post = DemoPostORM(title="Welcome", user=user)
        post.comments.append(DemoCommentORM(body="First"))
        user.secrets.append(DemoSecretORM(code_name="vault"))
        session.add(user)
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["can_delete"] is True
        assert payload["summary"] == {"roots": 1, "delete": 3, "protect": 0, "set_null": 0, "total": 4}
        child_model_slugs = [child["model_slug"] for child in payload["roots"][0]["children"]]
        assert "posts" in child_model_slugs
        assert None in child_model_slugs


async def test_router_uses_fallback_label_for_unregistered_related_model_in_delete_preview() -> None:
    app, session_factory = await build_test_app()

    async with session_factory() as session:
        user = DemoUserORM(username="alpha", password="hashed::1")
        user.secrets.append(DemoSecretORM(code_name="vault"))
        session.add(user)
        await session.commit()

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/1/delete-preview/")

        assert response.status_code == 200
        payload = response.json()
        secret_node = payload["roots"][0]["children"][0]
        assert secret_node["model_slug"] is None
        assert secret_node["model_title"] == "DemoSecretORM"
        assert secret_node["label"] == "vault"



from __future__ import annotations

from typing import Any

from fastapi import HTTPException, status
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect as sa_inspect

from xladmin.config import ModelConfig
from xladmin.introspection import convert_value_for_column, get_column_names, get_create_fields, get_update_fields
from xladmin.router_queries import convert_pk, resolve_relation_model


async def apply_payload_to_item(
        session: AsyncSession,
        model_config: ModelConfig,
        item: Any,
        payload: dict[str, Any],
        *,
        mode: str,
) -> None:
    mapper = sa_inspect(model_config.model)
    column_names = set(get_column_names(model_config))
    relationship_names = set(mapper.relationships.keys())
    allowed_fields = set(get_create_fields(model_config) if mode == "create" else get_update_fields(model_config))
    for field_name, raw_value in payload.items():
        if field_name not in allowed_fields:
            continue
        field_config = model_config.get_field_config(field_name)
        try:
            value = field_config.value_parser(raw_value) if field_config.value_parser is not None else raw_value

            if field_config.value_setter is not None:
                field_config.value_setter(item, value, payload, mode)
                continue

            if field_name in relationship_names:
                await assign_relationship_value(session, model_config, item, field_name, value)
                continue

            if field_name not in column_names:
                continue

            column = mapper.columns[field_name]
            if column.foreign_keys and value not in (None, ""):
                relation_model = resolve_relation_model(model_config, field_name)
                relation_pk_name = sa_inspect(relation_model).primary_key[0].key
                related_item = await session.get(relation_model, convert_pk(value))
                if related_item is None:
                    raise ValueError(f"Unknown related id for field '{field_name}'.")
                setattr(item, field_name, getattr(related_item, relation_pk_name))
                continue

            setattr(item, field_name, convert_value_for_column(column, value))
        except (TypeError, ValueError) as exc:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"Invalid value for field '{field_name}'.",
            ) from exc


async def assign_relationship_value(
        session: AsyncSession,
        model_config: ModelConfig,
        item: Any,
        field_name: str,
        value: Any,
) -> None:
    mapper = sa_inspect(model_config.model)
    relationship = mapper.relationships[field_name]
    relation_model = resolve_relation_model(model_config, field_name)
    relation_pk_name = sa_inspect(relation_model).primary_key[0].key

    if relationship.uselist:
        raw_ids = list(value or [])
        if not raw_ids:
            setattr(item, field_name, [])
            return
        normalized_ids = [convert_pk(item_id) for item_id in raw_ids]
        ordered_unique_ids = list(dict.fromkeys(normalized_ids))
        query = select(relation_model).where(getattr(relation_model, relation_pk_name).in_(ordered_unique_ids))
        related_items = list((await session.execute(query)).scalars().unique())
        related_items_by_pk = {
            getattr(related_item, relation_pk_name): related_item
            for related_item in related_items
        }
        if set(related_items_by_pk) != set(ordered_unique_ids):
            raise ValueError(f"Unknown related ids for field '{field_name}'.")
        setattr(item, field_name, [related_items_by_pk[item_id] for item_id in ordered_unique_ids])
        return

    if value in (None, ""):
        setattr(item, field_name, None)
        return

    related_item = await session.get(relation_model, convert_pk(value))
    if related_item is None:
        raise ValueError(f"Unknown related id for field '{field_name}'.")
    setattr(item, field_name, related_item)



from __future__ import annotations

import inspect
from typing import Any

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Response, status
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect as sa_inspect

from xladmin.config import HttpConfig, ModelConfig
from xladmin.delete_preview import build_delete_plan, build_delete_preview
from xladmin.i18n import translate
from xladmin.introspection import get_list_filter_input_kind, get_model_blocks_meta, get_model_meta
from xladmin.registry import build_registry
from xladmin.router_mutations import apply_payload_to_item
from xladmin.router_queries import (
    apply_list_filters,
    apply_ordering,
    apply_scoped_query,
    apply_search,
    build_model_query,
    convert_pk,
    count_total_items,
    get_item_by_pk,
    get_items_by_ids,
    resolve_label_field,
    resolve_list_filter_relation_model,
    resolve_relation_model,
)
from xladmin.router_schemas import (
    BulkActionPayload,
    ChoiceItemPayload,
    ChoicesResponse,
    DeletePreviewResponse,
    DeleteResultResponse,
    DetailResponse,
    FlexibleProcessedResponse,
    IdsPayload,
    ItemOnlyResponse,
    ItemPayload,
    ListResponse,
    ModelResponse,
    ModelsResponse,
    ObjectActionResponse,
    PaginationPayload,
)
from xladmin.serializer import serialize_model_instance


def create_router(config: HttpConfig) -> APIRouter:
    registry = build_registry(config.registry)
    router = APIRouter(prefix="/xladmin", tags=["XLAdmin"])
    model_meta_by_slug = {
        model_config.slug: get_model_meta(model_config, locale=registry.locale)
        for model_config in registry.list()
    }
    models_response = ModelsResponse(
        locale=registry.locale,
        items=list(model_meta_by_slug.values()),
        blocks=get_model_blocks_meta(registry),
    )

    def check_access(user: Any) -> None:
        if not config.is_allowed(user):
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=translate(registry.locale, "forbidden"))

    async def get_model_config(slug: str) -> ModelConfig:
        try:
            return registry.get(slug)
        except KeyError as exc:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "admin_model_not_found"),
            ) from exc

    async def get_existing_item(
            slug: str,
            item_id: str,
            session: AsyncSession,
            user: Any,
    ) -> tuple[ModelConfig, Any]:
        model_config = await get_model_config(slug)
        item = await get_item_by_pk(session, model_config, item_id, user, mode="detail")
        if item is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "object_not_found"),
            )
        return model_config, item

    async def execute_delete_plan(
            model_config: ModelConfig,
            items: list[Any],
            session: AsyncSession,
    ) -> int:
        preview, delete_items, set_null_items = await build_delete_plan(session, registry, model_config, items)
        if not preview["can_delete"]:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=translate(registry.locale, "delete_blocked"),
            )
        for related_item, attribute_names in set_null_items:
            for attribute_name in attribute_names:
                setattr(related_item, attribute_name, None)
        for item_to_delete in delete_items:
            await session.delete(item_to_delete)
        await session.commit()
        return len(items)

    @router.get("/models/", response_model=ModelsResponse)
    async def list_models(user: Any = Depends(config.get_current_user_dependency)) -> ModelsResponse:
        check_access(user)
        return models_response

    @router.get("/models/{slug}/", response_model=ModelResponse)
    async def get_model(
            slug: str,
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        check_access(user)
        try:
            return model_meta_by_slug[slug]
        except KeyError:
            await get_model_config(slug)
            raise

    @router.get("/models/{slug}/items/", response_model=ListResponse)
    async def list_items(
            slug: str,
            request: Request,
            limit: int = Query(default=50, ge=1, le=500),
            offset: int = Query(default=0, ge=0),
            q: str | None = None,
            sort: str | None = None,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ListResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        list_filter_values = {
            list_filter.slug: str(raw_value)
            for list_filter in model_config.list_filters
            if (raw_value := request.query_params.get(list_filter.slug)) not in (None, "")
        }
        query = await build_model_query(session, model_config, user, mode="list")
        query = apply_search(model_config, query, q, session)
        query = await apply_list_filters(model_config, query, list_filter_values, session, user)
        query = apply_ordering(model_config, query, sort)
        total = await count_total_items(session, model_config, query)
        items = list((await session.execute(query.limit(limit).offset(offset))).scalars().unique())
        return ListResponse(
            meta=model_meta_by_slug[model_config.slug],
            pagination=PaginationPayload(limit=limit, offset=offset, total=total),
            items=[serialize_model_instance(model_config, item, mode="list") for item in items],
        )

    @router.get("/models/{slug}/items/{item_id}/", response_model=DetailResponse)
    async def get_item(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> DetailResponse:
        check_access(user)
        model_config, item = await get_existing_item(slug, item_id, session, user)
        return DetailResponse(
            meta=model_meta_by_slug[model_config.slug],
            item=serialize_model_instance(model_config, item, mode="detail"),
        )

    @router.post("/models/{slug}/items/", status_code=status.HTTP_201_CREATED, response_model=ItemOnlyResponse)
    async def create_item(
            slug: str,
            payload: ItemPayload,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ItemOnlyResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        item = model_config.model()
        await apply_payload_to_item(session, model_config, item, payload.root, mode="create")
        session.add(item)
        await session.commit()
        await session.refresh(item)
        return ItemOnlyResponse(item=serialize_model_instance(model_config, item, mode="detail"))

    @router.patch("/models/{slug}/items/{item_id}/", response_model=ItemOnlyResponse)
    async def patch_item(
            slug: str,
            item_id: str,
            payload: ItemPayload,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ItemOnlyResponse:
        check_access(user)
        model_config, item = await get_existing_item(slug, item_id, session, user)
        await apply_payload_to_item(session, model_config, item, payload.root, mode="update")
        await session.commit()
        await session.refresh(item)
        refreshed_item = await get_item_by_pk(session, model_config, item_id, user, mode="detail")
        return ItemOnlyResponse(
            item=serialize_model_instance(model_config, refreshed_item or item, mode="detail"),
        )

    @router.get("/models/{slug}/items/{item_id}/delete-preview/", response_model=DeletePreviewResponse)
    async def get_delete_preview(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        check_access(user)
        model_config, item = await get_existing_item(slug, item_id, session, user)
        return await build_delete_preview(session, registry, model_config, [item])

    @router.delete("/models/{slug}/items/{item_id}/", status_code=status.HTTP_204_NO_CONTENT)
    async def delete_item(
            slug: str,
            item_id: str,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> Response:
        check_access(user)
        model_config, item = await get_existing_item(slug, item_id, session, user)
        await execute_delete_plan(model_config, [item], session)
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    @router.post(
        "/models/{slug}/items/{item_id}/actions/{action_slug}/",
        response_model=ObjectActionResponse,
    )
    async def object_action(
            slug: str,
            item_id: str,
            action_slug: str,
            payload: ItemPayload | None = None,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ObjectActionResponse:
        check_access(user)
        model_config, item = await get_existing_item(slug, item_id, session, user)

        action = next((item for item in model_config.object_actions if item.slug == action_slug), None)
        if action is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "object_action_not_found"),
            )

        result = action.handler(session, model_config, item, payload.root if payload is not None else {}, user)
        if inspect.isawaitable(result):
            result = await result
        await session.commit()
        await session.refresh(item)
        refreshed_item = await get_item_by_pk(session, model_config, item_id, user, mode="detail")
        return ObjectActionResponse(
            item=serialize_model_instance(model_config, refreshed_item or item, mode="detail"),
            result=result or {},
        )

    @router.post("/models/{slug}/bulk-delete/", response_model=DeleteResultResponse)
    async def bulk_delete(
            slug: str,
            payload: IdsPayload,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> DeleteResultResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        if not payload.ids:
            return DeleteResultResponse(deleted=0)
        items = await get_items_by_ids(session, model_config, payload.ids, user, mode="detail")
        deleted = await execute_delete_plan(model_config, items, session)
        return DeleteResultResponse(deleted=deleted)

    @router.post("/models/{slug}/bulk-delete-preview/", response_model=DeletePreviewResponse)
    async def bulk_delete_preview(
            slug: str,
            payload: IdsPayload,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> dict[str, Any]:
        check_access(user)
        model_config = await get_model_config(slug)
        if not payload.ids:
            return {
                "can_delete": True,
                "summary": {"roots": 0, "delete": 0, "protect": 0, "set_null": 0, "total": 0},
                "roots": [],
            }
        items = await get_items_by_ids(session, model_config, payload.ids, user, mode="detail")
        return await build_delete_preview(session, registry, model_config, items)

    @router.post("/models/{slug}/bulk-actions/{action_slug}/", response_model=FlexibleProcessedResponse)
    async def bulk_action(
            slug: str,
            action_slug: str,
            payload: BulkActionPayload,
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> FlexibleProcessedResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        if not payload.ids:
            return FlexibleProcessedResponse(processed=0)
        if action_slug == "delete":
            deleted_response = await bulk_delete(slug, IdsPayload(ids=list(payload.ids)), session, user)
            return FlexibleProcessedResponse(processed=deleted_response.deleted)

        action = next((item for item in model_config.bulk_actions if item.slug == action_slug), None)
        if action is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "bulk_action_not_found"),
            )

        items = await get_items_by_ids(session, model_config, list(payload.ids), user, mode="detail")
        result = action.handler(session, model_config, items, payload.action_payload, user)
        if inspect.isawaitable(result):
            result = await result
        await session.commit()
        response_payload = {"processed": len(items)}
        if result:
            response_payload.update(result)
        return FlexibleProcessedResponse.model_validate(response_payload)

    @router.get("/models/{slug}/fields/{field_name}/choices/", response_model=ChoicesResponse)
    async def field_choices(
            slug: str,
            field_name: str,
            q: str | None = None,
            ids: str | None = None,
            limit: int = Query(default=25, ge=1, le=100),
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ChoicesResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        relation_model = resolve_relation_model(model_config, field_name, locale=registry.locale)
        relation_config = registry.find_by_model(relation_model)
        label_field = (
                model_config.get_field_config(field_name).relation_label_field
                or resolve_label_field(relation_model)
        )
        return await load_relation_choices(
            session=session,
            relation_model=relation_model,
            relation_config=relation_config,
            label_field=label_field,
            q=q,
            ids=ids,
            limit=limit,
            user=user,
        )

    @router.get("/models/{slug}/filters/{filter_slug}/choices/", response_model=ChoicesResponse)
    async def filter_choices(
            slug: str,
            filter_slug: str,
            q: str | None = None,
            ids: str | None = None,
            limit: int = Query(default=25, ge=1, le=100),
            session: AsyncSession = Depends(config.get_db_session_dependency),
            user: Any = Depends(config.get_current_user_dependency),
    ) -> ChoicesResponse:
        check_access(user)
        model_config = await get_model_config(slug)
        list_filter = next((item for item in model_config.list_filters if item.slug == filter_slug), None)
        if list_filter is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=translate(registry.locale, "list_filter_not_found"),
            )

        input_kind = get_list_filter_input_kind(model_config, list_filter)
        if input_kind == "boolean":
            return ChoicesResponse(
                items=[
                    ChoiceItemPayload(id="true", label=translate(registry.locale, "yes")),
                    ChoiceItemPayload(id="false", label=translate(registry.locale, "no")),
                ],
            )
        if list_filter.options:
            return ChoicesResponse(
                items=[
                    ChoiceItemPayload(id=option.value, label=option.label)
                    for option in list_filter.options
                ],
            )

        relation_model = resolve_list_filter_relation_model(model_config, list_filter, locale=registry.locale)
        if relation_model is None:
            return ChoicesResponse(items=[])

        relation_config = registry.find_by_model(relation_model)
        label_field = list_filter.relation_label_field or resolve_label_field(relation_model)
        return await load_relation_choices(
            session=session,
            relation_model=relation_model,
            relation_config=relation_config,
            label_field=label_field,
            q=q,
            ids=ids,
            limit=limit,
            user=user,
        )

    return router


async def load_relation_choices(
        *,
        session: AsyncSession,
        relation_model: type[Any],
        relation_config: ModelConfig | None,
        label_field: str,
        q: str | None,
        ids: str | None,
        limit: int,
        user: Any,
) -> ChoicesResponse:
    relation_pk_name = sa_inspect(relation_model).primary_key[0].key
    relation_pk_attr = getattr(relation_model, relation_pk_name)
    selected_ids = [convert_pk(item_id) for item_id in ids.split(",") if item_id] if ids else []
    base_query = select(relation_model)
    if relation_config is not None:
        base_query = await apply_scoped_query(base_query, relation_config, session, user)

    selected_items = await fetch_choice_items(
        session=session,
        query=base_query.where(relation_pk_attr.in_(selected_ids)) if selected_ids else None,
    )
    search_query = base_query
    if q:
        if relation_config is not None:
            search_query = apply_search(relation_config, search_query, q, session)
        elif hasattr(relation_model, label_field):
            search_query = search_query.where(getattr(relation_model, label_field).ilike(f"%{q}%"))
    if selected_ids:
        search_query = search_query.where(~relation_pk_attr.in_(selected_ids))
    if hasattr(relation_model, label_field):
        search_query = search_query.order_by(getattr(relation_model, label_field).asc())
    search_items = await fetch_choice_items(
        session=session,
        query=search_query.limit(max(limit - len(selected_items), 0)),
    )
    return ChoicesResponse(
        items=[
            ChoiceItemPayload(
                id=getattr(item, relation_pk_name),
                label=str(getattr(item, label_field, None) or getattr(item, relation_pk_name)),
            )
            for item in [*selected_items, *search_items]
        ],
    )


async def fetch_choice_items(session: AsyncSession, query: Any | None) -> list[Any]:
    if query is None:
        return []
    return list((await session.execute(query)).scalars().unique())


create_admin_router = create_router



from __future__ import annotations

import pytest

from ._router_helpers import DemoUserORM, build_test_app, get_test_client, save_entities

pytestmark = pytest.mark.asyncio


async def test_router_returns_localized_forbidden_error() -> None:
    app, _session_factory = await build_test_app(locale="en", is_staff=False)

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/")

        assert response.status_code == 403
        assert response.json()["detail"] == "Access denied."


async def test_router_returns_localized_object_not_found_error() -> None:
    app, _session_factory = await build_test_app(locale="en")

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/999/delete-preview/")

        assert response.status_code == 404
        assert response.json()["detail"] == "Object was not found."


async def test_router_returns_localized_admin_model_not_found_error() -> None:
    app, _session_factory = await build_test_app(locale="en")

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/missing-model/")

        assert response.status_code == 404
        assert response.json()["detail"] == "Admin model was not found."


async def test_router_returns_localized_object_action_not_found_error() -> None:
    app, session_factory = await build_test_app(locale="en")
    await save_entities(session_factory, DemoUserORM(username="alpha", password="hashed::1"))

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/items/1/actions/missing-action/", json={})

        assert response.status_code == 404
        assert response.json()["detail"] == "Object action was not found."


async def test_router_returns_localized_bulk_action_not_found_error() -> None:
    app, session_factory = await build_test_app(locale="en")
    await save_entities(session_factory, DemoUserORM(username="alpha", password="hashed::1"))

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/bulk-actions/missing-action/", json={"ids": [1]})

        assert response.status_code == 404
        assert response.json()["detail"] == "Bulk action was not found."



from __future__ import annotations

from typing import Any

from sqlalchemy import inspect as sa_inspect
from sqlalchemy.orm import ColumnProperty, RelationshipProperty

from xladmin.config import AdminModelConfig


def get_mapper_field_names(config: AdminModelConfig) -> list[str]:
    mapper = sa_inspect(config.model)
    ordered_names: list[str] = []
    declared_names = getattr(config.model, "__annotations__", {})

    for field_name in declared_names:
        if field_name not in mapper.attrs:
            continue
        prop = mapper.attrs[field_name]
        if isinstance(prop, ColumnProperty | RelationshipProperty):
            ordered_names.append(prop.key)

    for prop in mapper.attrs:
        if not isinstance(prop, ColumnProperty | RelationshipProperty):
            continue
        if prop.key not in ordered_names:
            ordered_names.append(prop.key)

    return ordered_names


def get_column_names(config: AdminModelConfig) -> list[str]:
    mapper = sa_inspect(config.model)
    return [prop.key for prop in mapper.attrs if isinstance(prop, ColumnProperty)]


def get_relationship_names(config: AdminModelConfig) -> list[str]:
    mapper = sa_inspect(config.model)
    return [prop.key for prop in mapper.attrs if isinstance(prop, RelationshipProperty)]


def get_all_field_names(config: AdminModelConfig) -> list[str]:
    ordered_names = list(get_mapper_field_names(config))
    configured_names = get_configured_field_names(config)
    for field_name in config.fields:
        if field_name not in ordered_names:
            ordered_names.append(field_name)
    for field_name in configured_names:
        if field_name not in ordered_names:
            ordered_names.append(field_name)
    return ordered_names


def get_visible_list_fields(config: AdminModelConfig) -> list[str]:
    if config.list_display is not None:
        return list(config.list_display)
    if config.list_fields is not None:
        return list(config.list_fields)
    return [name for name in get_all_field_names(config) if not config.get_field_config(name).hidden_in_list]


def get_visible_detail_fields(config: AdminModelConfig) -> list[str]:
    if config.detail_fields is not None:
        return list(config.detail_fields)
    return [name for name in get_all_field_names(config) if not config.get_field_config(name).hidden_in_detail]


def get_create_fields(config: AdminModelConfig) -> list[str]:
    if config.create_fields is not None:
        return list(config.create_fields)
    mapper = sa_inspect(config.model)
    fields: list[str] = []
    column_names = set(get_column_names(config))
    relationship_names = set(get_relationship_names(config))
    for name in get_all_field_names(config):
        field_config = config.get_field_config(name)
        if is_read_only(config, name) or field_config.hidden_in_form:
            continue
        if name in column_names:
            column = mapper.columns[name]
            if name == get_pk_field_name(config) and pk_is_generated(column):
                continue
        elif name not in relationship_names and field_config.value_setter is None:
            continue
        fields.append(name)
    return fields


def get_update_fields(config: AdminModelConfig) -> list[str]:
    if config.update_fields is not None:
        return list(config.update_fields)
    relationship_names = set(get_relationship_names(config))
    column_names = set(get_column_names(config))
    return [
        name for name in get_all_field_names(config)
        if (
                name != get_pk_field_name(config)
                and not is_read_only(config, name)
                and not config.get_field_config(name).hidden_in_form
                and (
                        name in column_names
                        or name in relationship_names
                        or config.get_field_config(name).value_setter is not None
                )
        )
    ]


def get_pk_field_name(config: AdminModelConfig) -> str:
    mapper = sa_inspect(config.model)
    if hasattr(config.model, config.pk_field):
        return config.pk_field
    return mapper.primary_key[0].key


def get_pk_value(config: AdminModelConfig, instance: Any) -> Any:
    return getattr(instance, get_pk_field_name(config))


def get_display_value(config: AdminModelConfig, instance: Any) -> str:
    if config.display_field:
        value = getattr(instance, config.display_field, None)
        if value is not None:
            return str(value)
    pk_value = get_pk_value(config, instance)
    return f"{config.title} #{pk_value}"


def get_sort_column_name(config: AdminModelConfig, field_name: str) -> str | None:
    field_config = config.get_field_config(field_name)
    if field_config.ordering_field is not None:
        return field_config.ordering_field

    mapper = sa_inspect(config.model)
    if field_name in mapper.columns:
        return field_name

    if field_name in mapper.relationships and not mapper.relationships[field_name].uselist:
        relationship = mapper.relationships[field_name]
        local_columns = list(relationship.local_columns)
        if local_columns:
            return local_columns[0].key

    return None


def is_read_only(config: AdminModelConfig, field_name: str) -> bool:
    return config.get_field_config(field_name).read_only


def pk_is_generated(column: Any) -> bool:
    if not bool(column.primary_key):
        return False
    if column.autoincrement is True:
        return True
    if column.default is not None or column.server_default is not None:
        return True
    return False


def get_configured_field_names(config: AdminModelConfig) -> list[str]:
    names: list[str] = []
    for field_group in (
            config.list_display,
            config.list_fields,
            config.detail_fields,
            config.create_fields,
            config.update_fields,
    ):
        if field_group is None:
            continue
        for field_name in field_group:
            if field_name not in names:
                names.append(field_name)
    return names


def serialize_related_pk(instance: Any) -> Any:
    mapper = sa_inspect(type(instance))
    pk_column = mapper.primary_key[0]
    return getattr(instance, pk_column.key)
